#include "DistributeMat.h"

extern struct opts Options ;

void DistributeMatrixMondriaan(struct sparsematrix *pT, int P, int dir, 
			       double eps)
{
  int np ;  /* number of processors used in current partitioning */
  int p ;   /* p = P/np */
  int q ;   /* total number of splits to be done, P = 2^q */
  int r ;   /* number of splits currently done, np = 2^r  */
  int k, pl, pr ;
  double maxnz, e, d, growthfactor ;
  long t, nzAk, Ndumk, Ndiagk ;

  /*## Calculate q: ##*/
  q = 0 ; p = P ;
  while ( (p >>= 1)  > 0 )
    q++ ;

  maxnz = ((1 + eps) * (pT->Pstart[P] - pT->NrDummies)) / P   ;

  if ( Options.SplitStrategy == OneDimRow )
    dir = ROW ;

  if ( Options.SplitStrategy == OneDimCol )
    dir = COL ;

  if ( Options.SplitStrategy == Alternate && dir == Row )
    dir = ROW ;

  if ( Options.SplitStrategy == Alternate && dir == Col )
    dir = COL ;

  if ( Options.SplitStrategy == Alternate && dir == Ratio )
    {  
      if ( pT->m > pT->n )
        dir = ROW;
      else
        if ( pT->n > pT->m )
          dir = COL;
        else
          if ( (int) ( (2.0*rand()) / (RAND_MAX+1.0) ) == 0 )
            dir = ROW;
          else
            dir = COL;
    }

  d = 0.0 ;
  np = 1 ;
  p  = P ;
  r  = 0 ;
  growthfactor = 1.0 ;
  while ( p > 1 )
    {
#ifdef INFO2
      printf("************** Split %d -> %d parts **************\n", np, 2*np) ;
#endif

      if ( Options.LoadbalanceStrategy == Decrease )
        {
          if ( p==2 )
	    d = (1.0+eps) / growthfactor - 1.0 ;
          else
	    d = ( (1.0+eps) / growthfactor - 1.0 ) / 2 ;
          growthfactor *= (1.0+d) ;
        }
      else
	if ( Options.LoadbalanceStrategy == Increase )
          {
            if ( p==2 )
	      d = (1.0+eps) / growthfactor - 1.0 ;
            else
	      d = ( (1.0+eps) / growthfactor - 1.0 ) / p ;
            growthfactor *= (1.0+d) ;
          }
	else
          if ( Options.LoadbalanceStrategy == Dynamic )
            ; /* set d later, dependent on current partition */
          else
	    Abort("DistributeMatrixMondriaan(): "
	          "Unknown LoadbalanceStrategy") ;

      /*## Split all current partitions: ##*/
      for ( k = 0 ; k < np ; k++ )
	{
#ifdef INFO2
          printf("******** Split part %d ******** \n", k) ;
#endif
	  pl = k * p ;
	  pr = (k+1) * p ;

	  if ( Options.LoadbalanceStrategy == Dynamic )
	    {
	      nzAk = pT->Pstart[pr] - pT->Pstart[pl] ;
              Ndumk = 0 ;
              if ( pT->m == pT->n && pT->NrDummies > 0)
                { 
                  /* Count the dummies in part k */
                  for ( t = pT->Pstart[pl] ; t < pT->Pstart[pr]; t++ )
                    if (pT->i[t] == pT->j[t] && 
                        BSearchLong(pT->dummy, 0, pT->NrDummies-1, pT->i[t])
                                          != LONG_MIN )
                      Ndumk++ ;
                  nzAk -= Ndumk ;
                }

              if ( pT->m == pT->n &&
                   (pT->MMTypeCode[3]=='S' || pT->MMTypeCode[3]=='K' ||
                    pT->MMTypeCode[3]=='H') &&
                   Options.SymmetricMatrix_UseSingleEntry == SingleEntYes )
                {
                  /* Count the number of diagonal entries in part k */
                  Ndiagk = 0;
                  for ( t = pT->Pstart[pl] ; t < pT->Pstart[pr]; t++ )
                    if (pT->i[t] == pT->j[t])
                      Ndiagk++ ; 
                  if ( np==1) 
                    /* replace maxnz by weights: 2 for off-diagonal nonzeros,
                       1 for diagonal nondummy nonzeros, 0 for dummies */
                    maxnz = ((1 + eps) * (2*pT->Pstart[P] - Ndiagk - Ndumk))/P ;
                  nzAk = 2*nzAk - Ndiagk + Ndumk ; 
                }

              if ( nzAk != 0)
	        e =  (p * maxnz / nzAk ) - 1 ;
              else
                /* Every distribution is OK for dummies */
                e = 1.0 ; 
	      d = e / (q-r) ;
	    }

	  SplitMatrixKLFM(pT, pl, pr, dir, d) ;

	}

      /*## Alternate between split directions: ##*/
      if (Options.SplitStrategy == Alternate )
	dir = ! dir ;

#ifdef INFO2
      printf("Pstart = ") ;
      for (k = 0 ; k <= P ; k++)
      	printf("%6ld ", pT->Pstart[k]) ;
      printf("\n\n") ;
#endif      
  
      /*## Update p and np: ##*/
      np = np * 2 ;
      p  = p / 2 ;
      r++ ;
    }

  /*## Set matrix type code to distributed: ##*/
  pT->MMTypeCode[0] = 'D' ;

}  


void SplitMatrixKLFM(struct sparsematrix *pT, int pl, int pr, int dir, 
		       double eps) 
{
  int best ;
  struct sparsematrix A ;
  struct biparthypergraph HG, HGRow, HGCol ;

  if (pT->Pstart[pr] == pT->Pstart[pl])
    {
      /* matrix to be split is empty */
      pT->Pstart[(pl+pr)/2] = pT->Pstart[pl] ;
      return ;
    }

  /*## Copy info from T to A: ##*/
  A = *pT ;
  A.i = &(pT->i[pT->Pstart[pl]]) ;
  A.j = &(pT->j[pT->Pstart[pl]]) ;
  if (A.MMTypeCode[2] != 'P')
    A.ReValue = &(pT->ReValue[pT->Pstart[pl]]) ;
  if (A.MMTypeCode[2] == 'C')
    A.ImValue = &(pT->ImValue[pT->Pstart[pl]]) ;
  A.NrNzElts = pT->Pstart[pr] - pT->Pstart[pl] ;

  if ( Options.SplitStrategy == LocalBest )
    {
#ifdef INFO2
      printf("**** Try split in row direction \n") ;
#endif    
      HGRow = SparseMatrix2BiPartHyperGraph(A, ROW) ;
      RunMLGraphPart(&HGRow, eps) ;

#ifdef INFO2
      printf("**** Try split in col direction \n") ;
#endif   
      HGCol = SparseMatrix2BiPartHyperGraph(A, COL) ;
      RunMLGraphPart(&HGCol, eps) ;

#ifdef INFO2
      printf("Communication: row=%ld col=%ld\n", 
	      HGRow.OptComm, HGCol.OptComm) ;
#endif
      
      if ( HGRow.OptComm < HGCol.OptComm )
	best = ROW ;
      else
	if ( HGCol.OptComm < HGRow.OptComm )
	  best = COL ;
	else
	  if ( (int) ( (2.0*rand()) / (RAND_MAX+1.0) ) == 0 )
	    best = ROW ;
	  else
	    best = COL ;
      	    
#ifdef INFO2
      if ( best == ROW )
        printf("**** Row direction chosen\n") ;
      else
        printf("**** Col direction chosen\n") ;
#endif   

      if ( best == ROW )
	BiPartHyperGraph2SparseMatrix(HGRow, *pT, pl, pr) ;
      else
	BiPartHyperGraph2SparseMatrix(HGCol, *pT, pl, pr) ;

#ifdef WRITEISxJS
      if ( best == ROW )
	WriteISxJS(HGRow, pl, pr) ;
      else
	WriteISxJS(HGCol, pl, pr) ;      
#endif

      DeleteBiPartHyperGraph(HGRow) ;
      DeleteBiPartHyperGraph(HGCol) ;
    }
  else
    {
#ifdef INFO2
      if ( dir == ROW )
        printf("**** Split in row direction \n") ;
      else
        printf("**** Split in col direction \n") ;
#endif

      HG = SparseMatrix2BiPartHyperGraph(A, dir) ;
	
      RunMLGraphPart(&HG, eps) ;
	
      BiPartHyperGraph2SparseMatrix(HG, *pT, pl, pr) ;
	
#ifdef WRITEISxJS
      WriteISxJS(HG, pl, pr) ;
#endif
      DeleteBiPartHyperGraph(HG) ;
    }
}
