#ifndef __Graph_h__
#define __Graph_h__

#include "GainBucket.h"
#include "SparseMatrix.h"
#include "Sort.h"

/*## Bipartite Hyper Graph data structures : ##*/
typedef struct vertex
{
  long vtxwgt ;   /* weight of the vertex */
  long iStart ;   /* index into VtxAdjncy where the list begins */
  long iEnd   ;   /* index into VtxAdjncy where the list ends */
  int partition ; /* partition where this vertex belongs to (either 0 or 1) */
  struct bucketentry *GBentry ; /* pointer to the gainbucket */
} Vertex ;

typedef struct net
{
  long netwgt ;   /* weight of the net */
  long iStartP0 ; /* index into NetAdjncy where the list for partition 0 begins */
  long iStartP1 ; /* index into NetAdjncy where the list for partition 1 begins */
  long iEnd ;     /* index into NetAdjncy where the list ends */
} Net ;

typedef struct biparthypergraph
{
  long NrVertices ;
  long NrNets ;
  long NrEdges ;

  struct vertex *V ; /* stores the information about the vertices */
  long *VtxAdjncy ;  /* stores the adjacency lists of the vertices */

  struct net *N ;    /* stores the information about the nets */
  long *NetAdjncy ;  /* stores the adjacency lists of the nets */

  long *EdgeWeight ; /* stores the weights of the edges of the graph */         
                     /* in the same way as the adjacency list of the vertices */

  struct gainbucket GBVtx[2] ; /* pointers to the gainbuckets for the vertices */

  long WeightP[2] ;  /* stores the total weight for the (vertex) partitions */

  long *VtxMoveLog ; /* stores a log of the moved row vertices */
  long MinVtxLog, CurVtxLog ; /* indices for VtxMoveLog */

  long CurComm ; /* stores current communication */
  long MinComm ; /* stores minimum communication */
  long OptComm ; /* stores optimal communication (minimum over all iterations) */
  int *OptPartVtx ; /* stores the optimal partition for the Vertices */

  int SplitDir ; /* stores the split direction */

  /*## Matrix values/indices : (not used for a coarser hypergraph) ##*/
  double *MatReValue ; /* stores the real part of the values of the
                          nonzero matrix entries in the same way
                          as the adjacency list of the vertices */
  double *MatImValue ; /* stores the imaginary part of the values */
  long *Vtx2MatIndex ; /* stores the index of the matrix for the vertices */
  long *Net2MatIndex ; /* stores the index of the matrix for the nets */

} BiPartHyperGraph ;


typedef struct contraction
{
  long *Match ;
  long *Map ;
  long NrMatches ;
  long MaxNrVertices ; 
  long MaxVtxWgt ;
} Contraction ;



struct biparthypergraph CreateNewBiPartHyperGraph(long NrVertices, long NrNets,
						  long NrEdges, 
						  short StoreMat, char MType) ;

void DeleteBiPartHyperGraph(struct biparthypergraph HG) ;

struct biparthypergraph SparseMatrix2BiPartHyperGraph(struct sparsematrix A, 
						      int dir) ;

void BiPartHyperGraph2SparseMatrix(struct biparthypergraph HG, 
				   struct sparsematrix T, long pl, long pr) ;

void WriteISxJS(struct biparthypergraph HG, long pl, long pr) ;

#endif /* __Graph_h__ */
