#include "SparseMatrix.h"

void MMSparseMatrixInit(struct sparsematrix *pM)
{
  pM->MMTypeCode[0] = ' ' ;
  pM->MMTypeCode[1] = ' ' ;
  pM->MMTypeCode[2] = ' ' ;
  pM->MMTypeCode[3] = ' ' ;
  pM->Banner[0] = '\0' ;
  pM->Object[0] = '\0' ;
  pM->Format[0] = '\0' ;
  pM->Field[0] = '\0' ;
  pM->Symmetry[0] = '\0' ;
  pM->NrNzElts = 0 ;
  pM->m = 0 ;
  pM->n = 0 ;
  pM->i = NULL ;
  pM->j = NULL ;
  pM->NrDummies = 0 ;
  pM->NrProcs = 0 ;
  pM->ReValue = NULL ;
  pM->ImValue = NULL ;
  pM->dummy = NULL ;
  pM->Pstart = NULL ;
  pM->header = NULL ;
  pM->tail = NULL ;
}

struct sparsematrix MMReadSparseMatrix(char *filename) 
{
  struct sparsematrix M ;
  FILE *fp ;

  MMSparseMatrixInit(&M) ;
  
  /*## Open the filepointer: ##*/
  fp = OpenFile(filename, "r") ;

  /*## Read header: ##*/
  MMSparseMatrixReadHeader(&M, fp) ;

  /*## Allocate memory for the non-zero elements of the matrix: ##*/
  MMSparseMatrixAllocateMemory(&M) ;

  /*## Read the distribution information: ##*/
  MMSparseMatrixReadPstart(&M, fp) ;

  /*## Read the non-zero elements: ##*/
  MMSparseMatrixReadEntries(&M, fp) ;

  /*## Read the tail: ##*/
  MMSparseMatrixReadTail(&M, fp) ;

  /*## Close the file pointer: ##*/
  CloseFile(filename, fp) ;  

  return(M) ;
}

void MMPrintSparseMatrix(struct sparsematrix M, FILE *stream)
{
  /*## Print header: ##*/
  MMSparseMatrixPrintHeader(&M, stream) ;

  /*## Print the distribution information: ##*/
  MMSparseMatrixPrintPstart(&M, stream) ;

  /*## Print the non-zero elements: ##*/
  MMSparseMatrixPrintEntries(&M, stream) ;

  /*## Print the tail: ##*/
  MMSparseMatrixPrintTail(&M, stream) ;
}

void MMWriteSparseMatrix(struct sparsematrix M, char *filename)
{
  FILE *fp ;

  /*## Open the filepointer: ##*/
  fp = OpenFile(filename, "w") ;

  /*## Print matrix to filepointer: ##*/
  MMPrintSparseMatrix(M, fp) ;

  /*## Close the file pointer: ##*/
  CloseFile(filename, fp) ;
}

void MMDeleteSparseMatrix(struct sparsematrix *pM)
{
  MMSparseMatrixFreeMemory(pM) ;
  MMSparseMatrixInit(pM) ;
}


void MMSparseMatrixSetTypeCode(struct sparsematrix *pM, char *line) 
{
  char *c ;

  /*## Read Matrix Market Type Codes form first line: ##*/		
  if (sscanf(line, "%s %s %s %s %s", pM->Banner, pM->Object, 
	     pM->Format, pM->Field, pM->Symmetry) != 5 ||
      strcmp(pM->Banner, MM_Banner) )
    { 
      /*## No correct Type Code found, assume default: ##*/
      strcpy(pM->Banner, MM_Banner) ;
      pM->MMTypeCode[0] = 'M' ; strcpy(pM->Object, "matrix") ;
      pM->MMTypeCode[1] = 'C' ; strcpy(pM->Format, "coordinate") ;
      pM->MMTypeCode[2] = 'R' ; strcpy(pM->Field, "real") ;
      pM->MMTypeCode[3] = 'G' ; strcpy(pM->Symmetry, "general") ;
      return ;
    }

  /*## Set Object type: ##*/
  for ( c = pM->Object ; *c != '\0' ; c++)
    *c = tolower(*c) ;

  if ( ! strcmp(pM->Object, "matrix") )
    pM->MMTypeCode[0] = 'M' ;
  else
  if ( ! strcmp(pM->Object, "distributed-matrix") ||
       ! strcmp(pM->Object, "distributedmatrix") )
    pM->MMTypeCode[0] = 'D' ;      
  else
    {
      fprintf(stderr, "Unknown Object '%s'\n", pM->Object) ;
      Abort("MMSparseMatrixSetTypeCode(): incorrect type") ;
    }
  
  /*## Set Format type: ##*/
  for ( c = pM->Format ; *c != '\0' ; c++)
    *c = tolower(*c) ;

  if ( ! strcmp(pM->Format, "coordinate") || 
       ! strcmp(pM->Format, "sparse") )
    pM->MMTypeCode[1] = 'C' ;
  else
  if ( ! strcmp(pM->Format, "array") || 
       ! strcmp(pM->Format, "dense") )
    pM->MMTypeCode[1] = 'A' ;
  else
    {
      fprintf(stderr, "Unknown Format '%s'\n", pM->Format) ;
      Abort("MMSparseMatrixSetTypeCode(): incorrect type") ;
    }
  
  /*## Set Field type: ##*/
  for ( c = pM->Field ; *c != '\0' ; c++)
    *c = tolower(*c) ;

  if ( ! strcmp(pM->Field, "integer") )
    pM->MMTypeCode[2] = 'I' ;
  else
  if ( ! strcmp(pM->Field, "real") || 
       ! strcmp(pM->Field, "float") )
    pM->MMTypeCode[2] = 'R' ;
  else
  if ( ! strcmp(pM->Field, "complex") )
    pM->MMTypeCode[2] = 'C' ;
  else
  if ( ! strcmp(pM->Field, "pattern") )
    pM->MMTypeCode[2] = 'P' ;
  else
    {
      fprintf(stderr, "Unknown Field '%s'\n", pM->Field) ;
      Abort("MMSparseMatrixSetTypeCode(): incorrect type") ;
    }

  /*## Set Symmetry type: ##*/
  for ( c = pM->Symmetry ; *c != '\0' ; c++)
    *c = tolower(*c) ;

  if ( ! strcmp(pM->Symmetry, "general") )
    pM->MMTypeCode[3] = 'G' ;
  else
  if ( ! strcmp(pM->Symmetry, "symmetric") )
    pM->MMTypeCode[3] = 'S' ;
  else
  if ( ! strcmp(pM->Symmetry, "skew-symmetric") ||
       ! strcmp(pM->Symmetry, "skewsymmetric") )
    pM->MMTypeCode[3] = 'K' ;
  else
  if ( ! strcmp(pM->Symmetry, "hermitian") )
    pM->MMTypeCode[3] = 'H' ;
  else
    {
      fprintf(stderr, "Unknown Symmetry '%s'\n", pM->Symmetry) ;
      Abort("MMSparseMatrixSetTypeCode(): incorrect type") ;
    }
}

void MMSparseMatrixGetTypeCode(struct sparsematrix *pM, char *line) 
{
  /*## Set Banner: ##*/
  strcpy(pM->Banner, MM_Banner) ;

  /*## Set Object type: ##*/
  if ( pM->MMTypeCode[0] == 'M' )
    strcpy(pM->Object, "matrix") ;
  else
  if ( pM->MMTypeCode[0] == 'D' )
    strcpy(pM->Object, "distributed-matrix") ;
  else
    {
      fprintf(stderr, "Unknown Object Type Code '%c'\n", pM->MMTypeCode[0]) ;
      Abort("MMSparseMatrixGetTypeCode(): incorrect type code") ;
    }

  /*## Set Format type: ##*/
  if ( pM->MMTypeCode[1] == 'C' )
    strcpy(pM->Format, "coordinate") ;
  else
  if ( pM->MMTypeCode[1] == 'A' )
    strcpy(pM->Format, "array") ;
  else
    {
      fprintf(stderr, "Unknown Format Type Code '%c'\n", pM->MMTypeCode[1]) ;
      Abort("MMSparseMatrixGetTypeCode(): incorrect type code") ;
    }

  /*## Set Field type: ##*/
  if ( pM->MMTypeCode[2] == 'I' )
    strcpy (pM->Field, "integer") ;
  else
  if ( pM->MMTypeCode[2] == 'R' )
    strcpy(pM->Field, "real") ;
  else
  if ( pM->MMTypeCode[2] == 'C' )
    strcpy(pM->Field, "complex") ;
  else
  if ( pM->MMTypeCode[2] == 'P' )
    strcmp(pM->Field, "pattern") ;
  else
    {
      fprintf(stderr, "Unknown Field Type Code '%c'\n", pM->MMTypeCode[2]) ;
      Abort("MMSparseMatrixGetTypeCode(): incorrect type code") ;
    }

  /*## Set Symmetry type: ##*/
  if ( pM->MMTypeCode[3] == 'G' )
    strcpy(pM->Symmetry, "general") ;
  else
  if ( pM->MMTypeCode[3] == 'S' )
    strcpy(pM->Symmetry, "symmetric") ;    
  else
  if ( pM->MMTypeCode[3] == 'K' )
    strcpy(pM->Symmetry, "skew-symmetric")  ;
  else
  if ( pM->MMTypeCode[3] == 'H' )
    strcpy(pM->Symmetry, "hermitian") ;
  else
    {
      fprintf(stderr, "Unknown Symmetry Type Code '%c'\n", pM->MMTypeCode[3]) ;
      Abort("MMSparseMatrixGetTypeCode(): incorrect type code") ;
    }

  /*## Print the type codes to the header line: ##*/
  sprintf(line, "%s %s %s %s %s\n", pM->Banner, pM->Object, pM->Format, 
	  pM->Field, pM->Symmetry) ;
}


void MMSparseMatrixAllocateMemory(struct sparsematrix *pM) 
{
  if ( pM->NrNzElts > 0 )
    {
      /*## Allocate memory for indices: ##*/
      pM->i = (long *) malloc(pM->NrNzElts * sizeof(long)) ;
      pM->j = (long *) malloc(pM->NrNzElts * sizeof(long)) ;
      if ( pM->i == NULL || pM->j == NULL )
	Abort("MMSparseMatrixAllocateMemory(): Not enough memory "
	      "for matrix indices") ;

      /*## Allocate memory for numerical values: ##*/
      if ( pM->MMTypeCode[2] != 'P' )
	{
	  pM->ReValue = (double *) malloc(pM->NrNzElts * sizeof(double)) ;
	  if ( pM->ReValue == NULL )
	    Abort("MMSparseMatrixAllocateMemory(): Not enough memory "
		  "for matrix values") ;
	}
      if ( pM->MMTypeCode[2] == 'C' )
	{
	  pM->ImValue = (double *) malloc(pM->NrNzElts * sizeof(double)) ;
	  if ( pM->ImValue == NULL )
	    Abort("MMSparseMatrixAllocateMemory(): Not enough memory "
		  "for matrix values") ;
	}
    }

  /*## Allocate memory for Pstart: ##*/
  if ( pM->MMTypeCode[0] == 'D' && pM->NrProcs > 0 )
    {
      pM->Pstart = (long *) malloc((pM->NrProcs+1) * sizeof(long)) ;
      if ( pM->Pstart == NULL )
	Abort("MMSparseMatrixAllocateMemory(): Not enough memory "
	      "for Pstart") ;
    }
}

void MMSparseMatrixFreeMemory(struct sparsematrix *pM)
{
  if ( pM->NrNzElts > 0 )
    {
      if ( pM->i != NULL )
	free(pM->i) ;

      if ( pM->j != NULL )
	free(pM->j) ;

      if ( pM->MMTypeCode[2] != 'P' && pM->ReValue != NULL )
	free(pM->ReValue) ;
      
      if ( pM->MMTypeCode[2] == 'C' && pM->ImValue != NULL )
	free(pM->ImValue) ;
    }

  if ( pM->NrDummies > 0 && pM->dummy != NULL )
    free(pM->dummy) ;

  if ( pM->NrProcs > 0 && pM->Pstart != NULL )
    free(pM->Pstart) ;

  if ( pM->header != NULL )
    free(pM->header) ;

  if ( pM->tail != NULL )
    free(pM->tail) ;

} 


void MMSparseMatrixReadHeader(struct sparsematrix *pM, FILE *fp)
{
  int first, last, MatSize ;
  char *line, linebuffer[MM_MAX_LINE_LENGTH] ; 
  long hsize ;

  pM->MMTypeCode[0] = ' ' ;
  pM->header = NULL ;
  hsize = 1 ;
  MatSize = 0 ;

  while ( !MatSize && 
	  (line = fgets(linebuffer, MM_MAX_LINE_LENGTH, fp)) != NULL )
    { 
      first = 0 ; last = strlen(line) ;

      /*## Remove initial blank characters from the line: ##*/
      while ( isspace((int) linebuffer[first]) ) 
	first++ ;

      line = linebuffer + first ;

      if ( first != last )
	{
	  if ( linebuffer[first] == '%' )
	    {
	      if ( pM->MMTypeCode[0] == ' ' ) 
		MMSparseMatrixSetTypeCode(pM, line) ;
	      else
		{
		  /*## Add this line to the header: ##*/
		  hsize += last - first ;
		  pM->header = (char *) realloc(pM->header, hsize * 
						sizeof(char)) ;
		  if ( pM->header == NULL )
		    Abort("ReadHeader(): Not enough memory for realloc "
			  "of matrix header") ;
		  strncat(pM->header, line, last-first) ;
		}
	    }
	  else
	    {
	      /*## Read m, n, nnz, (P) from the first line: ##*/
	      sscanf(line, "%ld %ld %ld %d\n", &(pM->m), &(pM->n), 
		     &(pM->NrNzElts), &(pM->NrProcs)) ;
	      
	      if ( pM->m < 1 || pM->n < 1 || pM->NrNzElts < 1 || 
		   pM->NrProcs < 0 )
		{
		  CloseFile("", fp) ;
		  Abort("ReadHeader(): Error in matrix size or number of procs,"
                        "or matrix is empty") ;
		}
	      else
		MatSize = 1 ;
	    }
	}
    }

  /*## If there is no valid type code found, set it to default: ##*/
  if ( pM->MMTypeCode[0] == ' ' ) 
    MMSparseMatrixSetTypeCode(pM, "") ;
}

void MMSparseMatrixReadPstart(struct sparsematrix *pM, FILE *fp)
{
  register long t ;

  if ( pM->MMTypeCode[0] == 'D' )
    for ( t = 0 ; t <= pM->NrProcs ; t++  )
      if ( fscanf(fp, "%ld\n", &(pM->Pstart[t])) != 1 )
	Abort("MMSparseMatrixReadPstart(): Pstart: Read Error") ;
}

void MMSparseMatrixReadEntries(struct sparsematrix *pM, FILE *fp)
{
  register long t ;

  if ( pM->MMTypeCode[2] == 'P' )
    for ( t = 0 ; t < pM->NrNzElts ; t++ )
      {
	if ( fscanf(fp, "%ld %ld\n", &(pM->i[t]), &(pM->j[t])) != 2 )
	  Abort("MMSparseMatrixReadEntries(): Pattern: Read Error") ;
        if (pM->i[t] < 1 || pM->i[t] > pM->m || 
            pM->j[t] < 1 || pM->j[t] > pM->n )
	  Abort("MMSparseMatrixReadEntries(): Pattern: Read Error") ;
	pM->i[t]-- ; pM->j[t]-- ;
      }
  else
  if ( pM->MMTypeCode[2] == 'I' )
    for ( t = 0 ; t < pM->NrNzElts ; t++ )
      {
	if ( fscanf(fp, "%ld %ld %lg\n", &(pM->i[t]), &(pM->j[t]), 
		    &(pM->ReValue[t])) != 3 ) 
	  Abort("MMSparseMatrixReadEntries(): Integer: Read Error") ;
        if (pM->i[t] < 1 || pM->i[t] > pM->m ||  
            pM->j[t] < 1 || pM->j[t] > pM->n )
	  Abort("MMSparseMatrixReadEntries(): Integer: Read Error") ;
	pM->i[t]-- ; pM->j[t]-- ;
      }
  else
  if ( pM->MMTypeCode[2] == 'R' )
    for ( t = 0 ; t < pM->NrNzElts ; t++ )
      {
	if ( fscanf(fp, "%ld %ld %lg\n", &(pM->i[t]), &(pM->j[t]), 
		    &(pM->ReValue[t])) != 3 )
	  Abort("MMSparseMatrixReadEntries(): Real: Read Error") ; 
        if (pM->i[t] < 1 || pM->i[t] > pM->m ||  
            pM->j[t] < 1 || pM->j[t] > pM->n )
	  Abort("MMSparseMatrixReadEntries(): Real: Read Error") ; 
	pM->i[t]-- ; pM->j[t]-- ;
      }
  else
  if ( pM->MMTypeCode[2] == 'C' )
    for ( t = 0 ; t < pM->NrNzElts ; t++ )
      {
	if ( fscanf(fp, "%ld %ld %lg %lg\n", &(pM->i[t]), &(pM->j[t]), 
		    &(pM->ReValue[t]), &(pM->ImValue[t])) != 4 )
	  Abort("MMSparseMatrixReadEntries(): Complex: Read Error") ;
        if (pM->i[t] < 1 || pM->i[t] > pM->m ||  
            pM->j[t] < 1 || pM->j[t] > pM->n )
	  Abort("MMSparseMatrixReadEntries(): Complex: Read Error") ;
	pM->i[t]-- ; pM->j[t]-- ;
      }

  /* Check if symmetric matrix has entries only on or below the diagonal */
  if ( pM->MMTypeCode[3] == 'S' )
    for ( t = 0 ; t < pM->NrNzElts ; t++ )
      if (pM->i[t] < pM->j[t] )
        Abort("MMSparseMatrixReadEntries(): Symmetric: Read Error") ;

}

void MMSparseMatrixReadTail(struct sparsematrix *pM, FILE *fp)
{
  int first, last ;
  char *line, linebuffer[MM_MAX_LINE_LENGTH] ; 
  long tsize ;

  pM->tail = NULL ;
  tsize = 1 ;
  while ( (line = fgets(linebuffer, MM_MAX_LINE_LENGTH, fp)) != NULL )
    { 
      first = 0 ; last = strlen(line) ;

      /*## Remove initial blank characters from the line: ##*/
      while ( isspace((int) linebuffer[first]) ) 
	first++ ;

      line = linebuffer + first ;
      
      if ( first != last )
	{
	  /*## Add this line to the tail: ##*/
	  tsize += last - first ;
	  pM->tail = (char *) realloc(pM->tail, tsize * sizeof(char)) ; 
	  if ( pM->tail == NULL )
	    Abort("MMSparseMatrixReadTail(): Not enough memory for realloc "
		  "of matrix tail") ;
	  strncat(pM->tail, line, last-first) ;
	}
    }
}


void MMSparseMatrixPrintHeader(struct sparsematrix *pM, FILE *stream) 
{
  char line[MM_MAX_LINE_LENGTH] ; 

  /*## Print Matrix Market Type Codes to first line: ##*/
  MMSparseMatrixGetTypeCode(pM, line) ;
  fprintf(stream, "%s", line) ;

  /*## Print the header: ##*/
  if ( pM->header != NULL )
    fprintf(stream, "%s", pM->header) ;

  /*## Print m, n, nnz, (P) to the first line: ##*/
  if ( pM->MMTypeCode[0] == 'D' )
    fprintf(stream, "%ld %ld %ld %d\n", pM->m, pM->n, pM->NrNzElts, 
	    pM->NrProcs) ;
  else
    fprintf(stream, "%ld %ld %ld\n", pM->m, pM->n, pM->NrNzElts) ;
}

void MMSparseMatrixPrintPstart(struct sparsematrix *pM, FILE *stream)
{
  register long t ;

  if ( pM->MMTypeCode[0] == 'D' ) 
    for ( t = 0 ; t <= pM->NrProcs ; t++  )
      fprintf(stream, "%ld\n", pM->Pstart[t]) ;
}

void MMSparseMatrixPrintEntries(struct sparsematrix *pM, FILE *stream)
{
  register long t ;

  if ( pM->MMTypeCode[2] == 'P' )
    for ( t = 0 ; t < pM->NrNzElts ; t++ )
      fprintf(stream, "%ld %ld\n", pM->i[t]+1, pM->j[t]+1) ;
  else
  if ( pM->MMTypeCode[2] == 'I' )
    for ( t = 0 ; t < pM->NrNzElts ; t++ )
      fprintf(stream, "%ld %ld %ld\n", pM->i[t]+1, pM->j[t]+1, 
	      (long) pM->ReValue[t]) ; 
  else
  if ( pM->MMTypeCode[2] == 'R' )
    for ( t = 0 ; t < pM->NrNzElts ; t++ )
      fprintf(stream, "%ld %ld %g\n", pM->i[t]+1, pM->j[t]+1, pM->ReValue[t]) ;
  else
  if ( pM->MMTypeCode[2] == 'C' )
    for ( t = 0 ; t < pM->NrNzElts ; t++ )
      fprintf(stream, "%ld %ld %g %g\n", pM->i[t]+1, pM->j[t]+1, 
	      pM->ReValue[t], pM->ImValue[t]) ;
}

void MMSparseMatrixPrintTail(struct sparsematrix *pM, FILE *stream)
{
  /*## Print the tail: ##*/
  if ( pM->tail != NULL )
    fprintf(stream, "%s", pM->tail) ; 
}




void AddDummiesToSparseMatrix(struct sparsematrix *pM) 

/* Adds dummies to the square matrix to make its diagonal nonzero.
   This function should only be used before the partitioning,
   because it does not preserve distribution information. */

{
  register long t, tt ;
  short *MatrixDiagonal ;
  long l ;

  if ( pM->m != pM->n )
    Abort("AddDummiesToSparseMatrix(): Matrix is not square !") ;

  l = pM->n ;
  pM->NrDummies = l ;
  
  /*## Allocate memory for matrix diagonal: ##*/
  MatrixDiagonal = (short *) malloc(l * sizeof(short)) ;
  if ( MatrixDiagonal == NULL )
    Abort("AddDummiesToSparseMatrix(): Not enough memory for diagonal") ;

  /*## Init Diagonal: ##*/
  for ( t = 0 ; t < l ; t++ )
    MatrixDiagonal[t] = 0 ;

  /*## Fill the diagonal and count the number of dummies: ##*/
  for ( t = 0 ; t < pM->NrNzElts ; t++ )
    if ( pM->i[t] == pM->j[t] )
      if ( MatrixDiagonal[pM->i[t]] == 0 )
	{
	  MatrixDiagonal[pM->i[t]] = 1 ;
	  pM->NrDummies-- ;
	}

  /*## Allocate memory for dummy list: ##*/
  pM->dummy = (long *) malloc(pM->NrDummies * sizeof(long)) ;
  if ( pM->dummy == NULL )
    Abort("AddDummiesToSparseMatrix(): Not enough memory for dummy list") ;

  /*## Allocate memory for dummies: ##*/
  pM->i = (long *) realloc(pM->i, (pM->NrNzElts + pM->NrDummies) * 
			   sizeof(long)) ;
  pM->j = (long *) realloc(pM->j, (pM->NrNzElts + pM->NrDummies) * 
			   sizeof(long)) ;
  if ( pM->i == NULL || pM->j == NULL )
    Abort("AddDummiesToSparseMatrix(): Not enough memory for dummies");
  if ( pM->MMTypeCode[2] != 'P' )
    {
      pM->ReValue = (double *) realloc(pM->ReValue,
                        (pM->NrNzElts + pM->NrDummies) * sizeof(double)) ;
      if ( pM->ReValue == NULL )
        Abort("AddDummiesToSparseMatrix(): Not enough memory for dummies");
    }
  if ( pM->MMTypeCode[2] == 'C' )
    {
      pM->ImValue = (double *) realloc(pM->ImValue,
                        (pM->NrNzElts + pM->NrDummies) * sizeof(double)) ;
      if ( pM->ImValue == NULL )
        Abort("AddDummiesToSparseMatrix(): Not enough memory for dummies");
    }

  /*## Fill the dummy list: ##*/
  for ( t = 0, tt = 0 ; t < l ; t++ )
    if ( MatrixDiagonal[t] == 0 )
      pM->dummy[tt++] = t ;

  /*## Add the dummies to the matrix: ##*/
  for ( t = 0, tt = pM->NrNzElts ; t < pM->NrDummies ; t++, tt++ )
    {
      pM->i[tt] = pM->dummy[t] ;
      pM->j[tt] = pM->dummy[t] ;
      if ( pM->MMTypeCode[2] != 'P' )
        pM->ReValue[tt] = -1 ;
      if ( pM->MMTypeCode[2] == 'C' )
        pM->ImValue[tt] = -1 ; 
    }

  /*## Update the number of nonempty elements of the matrix: ##*/
  pM->NrNzElts += pM->NrDummies ;
}



void RemoveDummiesFromSparseMatrix(struct sparsematrix *pM) 

/* Removes the dummies and adjust the processor distribution
   (Pstart) accordingly, if needed. 
   This function can be used before and after partitioning. */

{
  register long t, tt, k ;
  int q ;
  long tmp, NrRemoved, start ;
  double dtmp ;
  
  /*## If the matrix contains no dummies, return ##*/
  if ( pM->NrDummies <= 0 )
    return ;

  /*## If the matrix is not square, there should be no dummies ! ##*/
  if ( pM->m != pM->n )
    Abort("RemoveDummiesFromSparseMatrix(): Matrix is not square") ;

  if ( pM->NrProcs < 1 || pM->Pstart == NULL )
    for ( t = 0, NrRemoved = 0 ; t < pM->NrNzElts ; t++ )
      while ( (k = pM->i[t]) == pM->j[t] && NrRemoved < pM->NrDummies &&
	      BSearchLong(pM->dummy, 0, pM->NrDummies-1, k) != LONG_MIN )
	{ 
	  tt = pM->NrNzElts-1 ;
	  
	  /*## Swap: pM->*[t] <-> pM->*[tt] ##*/	  
	  tmp = pM->i[t] ;
	  pM->i[t] = pM->i[tt] ;
	  pM->i[tt] = tmp ;
	  
	  tmp = pM->j[t] ;
	  pM->j[t] = pM->j[tt] ;
	  pM->j[tt] = tmp ;
	  
          if ( pM->MMTypeCode[2] != 'P' )
            {
	      dtmp = pM->ReValue[t] ;
	      pM->ReValue[t] = pM->ReValue[tt] ;
	      pM->ReValue[tt] = dtmp ;
            }

          if ( pM->MMTypeCode[2] == 'C' )
            {
              dtmp = pM->ImValue[t] ;
              pM->ImValue[t] = pM->ImValue[tt] ;
              pM->ImValue[tt] = dtmp ;
            }

	  pM->NrNzElts-- ;
	  NrRemoved++ ;
	}
  else
    {
      NrRemoved = 0 ;
      for ( q = 0, tt = 0, start = pM->Pstart[0] ; q < pM->NrProcs ; q++ )
	{
	  for ( t = start ; t < pM->Pstart[q+1] ; t++ )	  
	    {
	      /*## move: pM->*[t] --> pM->*[tt] ##*/
	      pM->i[tt] = pM->i[t] ;
	      pM->j[tt] = pM->j[t] ;
              if ( pM->MMTypeCode[2] != 'P' )
	        pM->ReValue[tt] = pM->ReValue[t] ;
              if ( pM->MMTypeCode[2] == 'C' )
	        pM->ImValue[tt] = pM->ImValue[t] ;
	      
	      if ( (k = pM->i[t]) == pM->j[t] &&
		   BSearchLong(pM->dummy, 0, pM->NrDummies-1, k) != LONG_MIN )
		NrRemoved++ ;
	      else
		tt++ ;
	      
	    }	
	  
	  /*## Update Pstart: ##*/
	  start = pM->Pstart[q+1] ;
	  pM->Pstart[q+1] = tt ;
	}

      /*## Update NrNzElts: ##*/
      pM->NrNzElts -= NrRemoved ;
    }

  /*## Check the number of removed dummies: ##*/
  if ( pM->NrDummies != NrRemoved ) 
    Abort("RemoveDummiesFromSparseMatrix(): Internal error: "
	  "NrDummies != NrRemoved ") ;


  /*## Free the memory of the dummies: ##*/
  pM->i = (long *) realloc(pM->i, pM->NrNzElts * sizeof(long)) ;
  pM->j = (long *) realloc(pM->j, pM->NrNzElts * sizeof(long)) ;
  if ( pM->i == NULL || pM->j == NULL )
    Abort("RemoveDummiesFromSparseMatrix(): Not enough memory") ;
  if ( pM->MMTypeCode[2] != 'P' )
    {
      pM->ReValue = (double *) realloc(pM->ReValue, pM->NrNzElts * 
				   sizeof(double)) ;
      if ( pM->ReValue == NULL )
        Abort("RemoveDummiesFromSparseMatrix(): Not enough memory") ;
    }
  if ( pM->MMTypeCode[2] == 'C' )
    {
      pM->ImValue = (double *) realloc(pM->ImValue, pM->NrNzElts * 
				   sizeof(double)) ;
      if ( pM->ImValue == NULL )
        Abort("RemoveDummiesFromSparseMatrix(): Not enough memory") ;
    } 


  if ( pM->NrDummies > 0 && pM->dummy != NULL )
      free(pM->dummy) ;
  pM->NrDummies = 0 ;
}

void SparseMatrixSymmetric2Full(struct sparsematrix *pM) 
{
/* This function can be used before and after partitioning,
   because it preserves the distribution information.
   It converts symmetric, skew-symmetric, and hermitian matrices
   to the full representation (which contains all nonzeros, not only
   those below or on the diagonal). */

  register long t, tt, Ndiag ;
  long start, NrNzEltsNew ;
  int q ;

  /*## Check if matrix is symmetric: ##*/
  if ( pM->m != pM->n || 
       (pM->MMTypeCode[3] != 'S' && pM->MMTypeCode[3] != 'K' &&
        pM->MMTypeCode[3] != 'H' ))
    Abort("SparseMatrixSymmetric2Full(): matrix is not symmetric") ;

  if ( pM->NrProcs < 1 || pM->Pstart == NULL )
    {
      /*## Count the number of diagonal elements: ##*/
      for ( t = 0, Ndiag = 0 ; t < pM->NrNzElts ; t++ )
        if (pM->i[t] == pM->j[t])
          Ndiag++ ;
      NrNzEltsNew = 2 * pM->NrNzElts - Ndiag;
      if ( pM->MMTypeCode[3] == 'K' && Ndiag > 0 )
        Abort("SparseMatrixSymmetric2Full(): matrix is not skew-symmetric") ;
    }
  else
    {
      /*## Update Pstart: ##*/
      for ( q = 0, tt = 0, start = pM->Pstart[0] ; q < pM->NrProcs ; q++ )
        {
          for ( t = start ; t < pM->Pstart[q+1] ; t++ )
            {
              if ( pM->i[t] == pM->j[t] )
                tt++ ;
              else
                tt += 2;
            }    
         
          start = pM->Pstart[q+1] ;
          pM->Pstart[q+1] = tt ;
        }
      NrNzEltsNew = tt;
    }
 
  /*## Allocate memory for the entries: ##*/
  pM->i = (long *) realloc(pM->i, NrNzEltsNew * sizeof(long)) ;
  pM->j = (long *) realloc(pM->j, NrNzEltsNew * sizeof(long)) ;
  if ( pM->i == NULL || pM->j == NULL )
    Abort("SparseMatrixSymmetric2Full(): Not enough memory to store"
          "full matrix") ;
  if ( pM->MMTypeCode[2] != 'P' )
    {
      pM->ReValue = (double *) realloc(pM->ReValue, NrNzEltsNew * 
				   sizeof(double)) ;
      if ( pM->ReValue == NULL )
        Abort("SparseMatrixSymmetric2Full(): Not enough memory to store"
              "full matrix") ;
    }
  if ( pM->MMTypeCode[2] == 'C' )
    {
      pM->ImValue = (double *) realloc(pM->ImValue, NrNzEltsNew * 
				   sizeof(double)) ;
      if ( pM->ImValue == NULL )
        Abort("SparseMatrixSymmetric2Full(): Not enough memory to store"
              "full matrix") ;
    }

  /*## Copy the entries: ##*/
  for ( t = pM->NrNzElts - 1, tt = NrNzEltsNew -1 ; t >= 0; t-- )
    {
      /* Copy entry t into tt */
      pM->i[tt] = pM->i[t] ;
      pM->j[tt] = pM->j[t] ;
      if ( pM->MMTypeCode[2] != 'P' )
        pM->ReValue[tt] = pM->ReValue[t] ;
      if ( pM->MMTypeCode[2] == 'C' )
        pM->ImValue[tt] = pM->ImValue[t] ;
      tt-- ;   

      /* Copy and transpose nondiagonal entry t */
      if (pM->i[tt+1] != pM->j[tt+1])
        {
          pM->i[tt] = pM->j[tt+1] ; 
          pM->j[tt] = pM->i[tt+1] ; 
          if ( pM->MMTypeCode[2] != 'P' )
            {
              if ( pM->MMTypeCode[3] == 'K' )
                pM->ReValue[tt] = -pM->ReValue[tt+1] ; 
              else
                pM->ReValue[tt] = pM->ReValue[tt+1] ; 
            }
          if ( pM->MMTypeCode[2] == 'C' )
            { 
              if ( pM->MMTypeCode[3] == 'K' || pM->MMTypeCode[3] == 'H' ) 
                pM->ImValue[tt] = -pM->ImValue[tt+1] ;
              else 
                pM->ImValue[tt] = pM->ImValue[tt+1] ;
            } 
          tt-- ;
        }
    }

  /*## Update the number of nonzero entries and the matrix type code: ##*/
  pM->NrNzElts = NrNzEltsNew;
  pM->MMTypeCode[3] = 'G' ; strcpy(pM->Symmetry, "general") ;
 
}

void SparseMatrixSymmetricLower2Random(struct sparsematrix *pM)
/* This function randomly transposes the indices of every nonzero
   with a probablity of 50 percent, without changing the numerical value.
   This function preserves the distribution information ! */
{
  register long t, tmp ;
  register int c ;

  /*## Check if matrix is square and symmetric: ##*/
  if ( pM->m != pM->n || 
       (pM->MMTypeCode[3] != 'S'  && pM->MMTypeCode[3] != 'K'  &&
        pM->MMTypeCode[3] != 'H' ) ) 
    Abort("SparseMatrixSymmetricLower2Random(): matrix is not symmetric") ;
  
  for ( t = 0 ; t < pM->NrNzElts ; t++ )	  
    {
      c = (int) ( (2.0*rand()) / (RAND_MAX+1.0) ) ;
      if ( c == 0 )
	{
	  /*## Swap index i and j: ##*/
	  tmp = pM->i[t] ;
	  pM->i[t] = pM->j[t] ;
	  pM->j[t] = tmp ;
	}
    }
}

void SparseMatrixSymmetricRandom2Lower(struct sparsematrix *pM)
{
  /* This function preserves the distribution information ! */
  register long t, tmp ;

  /*## Check if matrix is square and symmetric: ##*/
  if ( pM->m != pM->n || 
       (pM->MMTypeCode[3] != 'S'  && pM->MMTypeCode[3] != 'K'  &&
        pM->MMTypeCode[3] != 'H' ) )   
    Abort("SparseMatrixSymmetricRandom2Lower(): matrix is not symmetric") ;
 
  for ( t = 0 ; t < pM->NrNzElts ; t++ )
    {
      if ( pM->i[t] < pM->j[t] )
        {
          /*## Swap index i and j: ##*/
          tmp = pM->i[t] ;
          pM->i[t] = pM->j[t] ;
          pM->j[t] = tmp ;
        }
    }
}

