#ifndef __SparseMatrix_h__
#define __SparseMatrix_h__

#include "IO.h"
#include "Sort.h"

#define MM_Banner           "%%MatrixMarket"
#define MM_MAX_LINE_LENGTH  1025
#define MM_MAX_WORD_LENGTH  129

typedef struct sparsematrix
{
  char MMTypeCode[4] ; 
  char Banner[MM_MAX_WORD_LENGTH] ;
  char Object[MM_MAX_WORD_LENGTH] ;
  char Format[MM_MAX_WORD_LENGTH] ;
  char Field[MM_MAX_WORD_LENGTH] ;
  char Symmetry[MM_MAX_WORD_LENGTH] ;

  long NrNzElts ;
  long m ;
  long n ;

  long NrDummies ;
  int  NrProcs ;

  long *i ; /* List of row indices */
  long *j ; /* List of column indices */
 
  double *ReValue ; /* List of Real part of the matrix values */
  double *ImValue ; /* List of Imaginary part of the matrix values */
  
  long *dummy ;  /* List of dummy elements */
  long *Pstart ; /* Pointer to the start of the triples for processor P */

  char *header ; /* header comment of the matrix */
  char *tail ;   /* tail comment of the matrix */
} SparseMatrix ;


void MMSparseMatrixInit(struct sparsematrix *pM) ;
struct sparsematrix MMReadSparseMatrix(char *filename) ;
void MMPrintSparseMatrix(struct sparsematrix M, FILE *stream) ;
void MMWriteSparseMatrix(struct sparsematrix M, char *filename) ;
void MMDeleteSparseMatrix(struct sparsematrix *pM) ;
void AddDummiesToSparseMatrix(struct sparsematrix *pM) ;
void RemoveDummiesFromSparseMatrix(struct sparsematrix *pM) ;
void SparseMatrixSymmetric2Full(struct sparsematrix *pM) ;
void SparseMatrixSymmetricLower2Random(struct sparsematrix *pM) ;
void SparseMatrixSymmetricRandom2Lower(struct sparsematrix *pM) ;

void MMSparseMatrixSetTypeCode(struct sparsematrix *pM, char *line) ;
void MMSparseMatrixGetTypeCode(struct sparsematrix *pM, char *line) ;
void MMSparseMatrixAllocateMemory(struct sparsematrix *pM) ;
void MMSparseMatrixFreeMemory(struct sparsematrix *pM) ;
void MMSparseMatrixReadHeader(struct sparsematrix *pM, FILE *fp) ;
void MMSparseMatrixReadPstart(struct sparsematrix *pM, FILE *fp) ;
void MMSparseMatrixReadEntries(struct sparsematrix *pM, FILE *fp) ;
void MMSparseMatrixReadTail(struct sparsematrix *pM, FILE *fp) ;
void MMSparseMatrixPrintHeader(struct sparsematrix *pM, FILE *stream) ;
void MMSparseMatrixPrintPstart(struct sparsematrix *pM, FILE *stream) ;
void MMSparseMatrixPrintEntries(struct sparsematrix *pM, FILE *stream) ;
void MMSparseMatrixPrintTail(struct sparsematrix *pM, FILE *stream) ;

#endif /* __SparseMatrix_h__ */
